package com.winhc.bigdata.task.framework.odps;

import com.winhc.bigdata.task.enums.DataWorksTaskFlowEnum;
import com.winhc.bigdata.task.enums.TaskStatusEnum;
import com.winhc.bigdata.task.framework.odps.entity.MaxComputeInstance;
import com.winhc.bigdata.task.framework.odps.entity.OdpsResultInfo;
import com.winhc.bigdata.task.framework.odps.entity.PullRequestRecord;
import com.winhc.bigdata.task.framework.odps.exception.AsyncTaskRunningException;
import com.winhc.bigdata.task.framework.odps.service.MaxComputeInstanceService;
import com.winhc.bigdata.task.framework.odps.service.OdpsResultInfoService;
import com.winhc.bigdata.task.framework.odps.service.PullRequestRecordService;
import com.winhc.bigdata.task.framework.odps.utils.ExecOdpsSqlUtil;
import com.winhc.bigdata.task.framework.odps.utils.ExportFlowUtils;
import com.winhc.bigdata.task.service.DataWorksService;
import com.winhc.bigdata.task.service.DingTalkService;
import com.winhc.bigdata.task.util.DataWorksFlowStatusUtil;
import lombok.AllArgsConstructor;
import lombok.SneakyThrows;
import lombok.extern.slf4j.Slf4j;
import org.springframework.scheduling.annotation.Scheduled;
import org.springframework.stereotype.Component;

import java.util.List;
import java.util.Map;

/**
 * @author: XuJiakai
 * 2021/7/6 10:24
 */
@Slf4j
@Component
@AllArgsConstructor
public class ExportDataCenter {
    private final PullRequestRecordService pullRequestRecordService;
    private final OdpsResultInfoService odpsResultInfoService;
    private final MaxComputeInstanceService maxComputeInstanceService;
    private final ExportFlowUtils exportFlowUtils;
    private final ExecOdpsSqlUtil execOdpsSqlUtil;
    private final DataWorksService dataWorksService;
    private final DingTalkService dingTalkService;

    private final AsyncTaskHandle asyncTaskHandle;

    /**
     * 定时判断是否有拉数据请求
     */
    @Scheduled(cron = "0 0/1 * * * ? ")
    public synchronized void pull() {
//        log.info("处理拉数据请求...");
        List<PullRequestRecord> list = pullRequestRecordService.lambdaQuery()
                .ne(PullRequestRecord::getPartitionTaskStatus, TaskStatusEnum.导出完毕)
                .isNull(PullRequestRecord::getPullInstanceId)
                .list();
        for (PullRequestRecord record : list) {
            String sqlId = record.getSqlId();
            String operatorName = record.getPullBy();
            log.info("拉取数据人：{} 拉取sqlId:{}", operatorName, sqlId);
            pullDetail(sqlId, operatorName);
        }
    }


    private void pullDetail(String sqlId, String operatorName) {
        List<OdpsResultInfo> list = odpsResultInfoService.lambdaQuery()
                .eq(OdpsResultInfo::getSqlId, sqlId)
                .eq(OdpsResultInfo::getPartitionTaskStatus, TaskStatusEnum.生成完毕)
                .orderByAsc(OdpsResultInfo::getPriorityLevel)
                .list();
        if (list.isEmpty()) {
            //执行一个sql
            try {
                asyncTaskHandle.run(sqlId, operatorName);
            } catch (AsyncTaskRunningException e) {
            }
        } else {
            //启动一个导数据流程
            OdpsResultInfo odpsResultInfo = list.get(0);
            PullRequestRecord one = pullRequestRecordService.lambdaQuery()
                    .eq(PullRequestRecord::getSqlId, sqlId)
                    .isNull(PullRequestRecord::getPullInstanceId)
                    .one();

            String exportInstanceId = exportFlowUtils.export(odpsResultInfo.getPartitionName()) + "";

            log.info("启动导数据任务：{} {}", odpsResultInfo.getTableName(), odpsResultInfo.getPartitionName());
            one.setPullInstanceId(exportInstanceId);
            one.setPartitionTaskStatus(TaskStatusEnum.导出中);
            one.setOdpsResultInfoId(odpsResultInfo.getId());
            pullRequestRecordService.updateById(one);
        }
    }


    /**
     * 定时判断sql是否执行完毕
     */
    @Scheduled(cron = "0 0/1 * * * ? ")
    public synchronized void sql() {
//        log.info("处理sql执行任务...");
        List<MaxComputeInstance> list = maxComputeInstanceService.lambdaQuery()
                .eq(MaxComputeInstance::getInstanceTerminated, false)
                .list();
        for (MaxComputeInstance instance : list) {
            asyncTaskHandle.isSuccess(instance.getId());
        }
    }


    /**
     * 定时判断导数据DI是否完成
     */
    @Scheduled(cron = "0 0/1 * * * ? ")
    @SneakyThrows
    public synchronized void export() {
//        log.info("处理导出数据任务...");
        List<PullRequestRecord> list = pullRequestRecordService.lambdaQuery()
                .ne(PullRequestRecord::getPartitionTaskStatus, TaskStatusEnum.导出完毕)
                .isNotNull(PullRequestRecord::getPullInstanceId)
                .list();

        for (PullRequestRecord record : list) {
            String pullInstanceId = record.getPullInstanceId();
            Map<String, DataWorksTaskFlowEnum> winhc_ng = dataWorksService.query("winhc_ng", Long.parseLong(pullInstanceId));
            Boolean successful = DataWorksFlowStatusUtil.isSuccessful(winhc_ng);

            if (successful != null) {
                if (successful) {
                    record.setPartitionTaskStatus(TaskStatusEnum.导出完毕);
                    odpsResultInfoService.lambdaUpdate()
                            .eq(OdpsResultInfo::getId, record.getOdpsResultInfoId())
                            .set(OdpsResultInfo::getPartitionTaskStatus, TaskStatusEnum.导出完毕)
                            .update();
                } else {
                    log.error("导出失败");
                    dingTalkService.error("导出数据失败，dagId:{}", pullInstanceId);
                    record.setPartitionTaskStatus(TaskStatusEnum.等待中);
                }
                pullRequestRecordService.updateById(record);
            }
        }
    }
}
