package com.winhc.bigdata.task.util;

import com.alibaba.fastjson.JSONObject;
import com.google.common.util.concurrent.ThreadFactoryBuilder;
import lombok.SneakyThrows;
import lombok.extern.slf4j.Slf4j;
import org.elasticsearch.action.search.*;
import org.elasticsearch.client.RequestOptions;
import org.elasticsearch.client.RestHighLevelClient;
import org.elasticsearch.common.unit.TimeValue;
import org.elasticsearch.index.query.QueryBuilder;
import org.elasticsearch.index.query.QueryBuilders;
import org.elasticsearch.index.query.WrapperQueryBuilder;
import org.elasticsearch.search.Scroll;
import org.elasticsearch.search.SearchHit;
import org.elasticsearch.search.builder.SearchSourceBuilder;

import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;
import java.util.function.Consumer;

/**
 * @author: XuJiakai
 * 2020/11/16 19:23
 */
@Slf4j
public class EsFastScan {
    private final Scroll scroll = new Scroll(TimeValue.timeValueMinutes(3L));

    private RestHighLevelClient client;
    private Consumer<SearchHit[]> func;
    private String index;
    private String type;
    private QueryBuilder queryBuilder;
    private int batchSize = 500;
    private int thread_num = 10;

    public EsFastScan(RestHighLevelClient client, Consumer<SearchHit[]> func, String index, String type) {
        this.client = client;
        this.func = func;
        this.index = index;
        this.type = type;
    }

    public EsFastScan(RestHighLevelClient client, Consumer<SearchHit[]> func, String index, String type, QueryBuilder queryBuilder) {
        this.client = client;
        this.func = func;
        this.index = index;
        this.type = type;
        this.queryBuilder = queryBuilder;
    }

    public EsFastScan(RestHighLevelClient client, Consumer<SearchHit[]> func, String index, String type, String dsl) {
        this.client = client;
        this.func = func;
        this.index = index;
        this.type = type;
        this.queryBuilder = getQuery(JSONObject.parseObject(dsl));
    }


    public EsFastScan batchSize(int batchSize) {
        this.batchSize = batchSize;
        return this;
    }


    @SneakyThrows
    public void scan() {
        int poolSize = thread_num;
        ArrayBlockingQueue<Runnable> objects = new ArrayBlockingQueue<>(poolSize * 4);
        ThreadPoolExecutor executorService = new ThreadPoolExecutor(
                poolSize, poolSize * 2,
                10L, TimeUnit.SECONDS,
                objects,
                new ThreadFactoryBuilder().setNameFormat("ScanEs-pool").build(),
                new ThreadPoolExecutor.CallerRunsPolicy()
        );


        SearchRequest searchRequest = new SearchRequest(index).types(type).scroll(scroll);

        SearchSourceBuilder searchSourceBuilder = new SearchSourceBuilder().size(batchSize);
        if (queryBuilder != null) {
            searchSourceBuilder.query(queryBuilder);
        }

        searchRequest.source(searchSourceBuilder);
        SearchResponse searchResponse = client.search(searchRequest);

        long totalHits = searchResponse.getHits().getTotalHits();
        int n = searchResponse.getHits().getHits().length;
        String scrollId = searchResponse.getScrollId();
        if (n != 0) {
            executorService.submit(() -> {
                log.info("执行一批：{}", n);
                func.accept(searchResponse.getHits().getHits());
            });
        } else {
            log.warn("数据为空！");
        }

        do {
            int num = scanData(executorService, scrollId);
            if (num == 0) {
                break;
            }
        } while (true);
        log.info("es scan is successful，total hits: {}", totalHits);
        try {
            log.info("scroll id: {}", scrollId);
            ClearScrollRequest clearScrollRequest = new ClearScrollRequest();
            clearScrollRequest.addScrollId(scrollId);
            client.clearScroll(clearScrollRequest, RequestOptions.DEFAULT);
            log.info("clean scroll id success !");
        } catch (Exception e) {
//            log.error(e.getMessage(), e);
        }
    }

    @SneakyThrows
    private int scanData(ThreadPoolExecutor executorService, String scrollId) {
        SearchScrollRequest scroll = new SearchScrollRequest(scrollId);
        SearchResponse searchResponse = client.scroll(scroll, RequestOptions.DEFAULT);
        int num = searchResponse.getHits().getHits().length;
        if (num == 0) {
            return num;
        }
        executorService.submit(() -> {
            log.info("执行一批：{}", num);
            func.accept(searchResponse.getHits().getHits());
        });
        return num;
    }


    private static final String q = "query";

    private static WrapperQueryBuilder getQuery(JSONObject jsonObject) {
        String dsl_json;
        if (jsonObject.containsKey(q)) {
            dsl_json = JsonUtils.jsonObjToString(jsonObject.getJSONObject(q));
        } else {
            dsl_json = JsonUtils.jsonObjToString(jsonObject);
        }
        return QueryBuilders.wrapperQuery(dsl_json);
    }
}
